import typing

if typing.TYPE_CHECKING:  # pragma: no cover
    from .applications import App


class SettingsError(Exception):
    """Raised when a setting is missing, ill-declared or invalid."""


class Settings:
    def __init__(self, obj: typing.Optional[typing.Any]):
        for setting in dir(obj):
            if not setting.isupper() or setting.startswith("_"):
                continue
            value = getattr(obj, setting)
            setattr(self, setting, value)


class LazySettings:
    """A lazy proxy for application settings.

    Once configured, an instance of this class can be used to access settings
    from anywhere in the application code base.

    Such an instance is in fact exposed as `bocadillo.settings`.

    Settings can be accessed using:

    - Dot notation: `settings.FOO`.
    - The `getattr` builtin: `getattr(settings, "FOO")`.
    - The dict-like `.get()` method: `settings.get("FOO", "foo")`.
    """

    def __init__(self):
        self._wrapped = None

    def configure(self, obj: typing.Any = None, **options):
        if self.configured:
            raise RuntimeError("Settings are already configured")

        wrapped = Settings(obj)

        for name, option in options.items():
            assert name.isupper()
            setattr(wrapped, name, option)

        self._wrapped = wrapped

    @property
    def configured(self) -> bool:
        return self._wrapped is not None

    def __getattr__(self, name: str) -> typing.Any:
        if not self.configured:
            raise SettingsError(
                f"Requested setting {name} but settings aren't configured yet."
            )

        value = getattr(self._wrapped, name)

        self.__dict__[name] = value  # cache setting

        return value

    def __setattr__(self, name: str, value: typing.Any):
        if name == "_wrapped":
            self.__dict__.clear()
        else:
            self.__dict__.pop(name, None)  # remove from cache
        super().__setattr__(name, value)

    def __contains__(self, name: str) -> bool:
        return name in self.__dict__

    def get(self, name: str, default: typing.Any = None) -> typing.Any:
        return getattr(self, name, default)

    def _clear(self):
        self._wrapped = None


settings = LazySettings()  # pylint: disable=invalid-name


def configure(app: "App", settings_obj: typing.Any = None, **kwargs) -> "App":
    """Configure the application settings and setup plugins.

    # Parameters
    app (App): an application instance.
    settings (any): an optional settings object or module.
    **kwargs (any): arbitrary settings, case-insensitive.

    # Returns
    app (App): the same object as `app`, for convenience.
    """
    from .plugins import setup_plugins

    if settings_obj is None:
        settings_obj = kwargs.pop("settings", None)
    kwargs = {key.upper(): value for key, value in kwargs.items()}
    settings.configure(obj=settings_obj, **kwargs)
    setup_plugins(app)

    return app



from invoke import task


@task
def install(c):
    c.run("pip install -r requirements.txt")
    c.run("pre-commit install")


@task
def test(c):
    c.run("pytest")


@task
def coverage(c, missing=False):
    command = "pytest --cov=bocadillo tests/ --cov-fail-under=100"
    if missing:
        command += " --cov-report term-missing --cov-report term:skip-covered"
    c.run(command)


@task
def ci(c):
    c.run("pip install codecov")
    coverage(c)


@task
def apiref(c):
    c.run("pydocmd generate")


@task
def bumpversion(c, level):
    c.run("pip install bumpversion")
    c.run(f"bash scripts/bumpversion.sh {level}")



import typing

import pytest

from bocadillo import App, configure, create_client, Templates, settings


@pytest.fixture(name="raw_app")
def fixture_raw_app(request) -> App:
    settings._clear()
    return App()


@pytest.fixture(name="app")
def fixture_app(raw_app: App) -> App:
    configure(raw_app)
    return raw_app


@pytest.fixture
def client(app):
    return create_client(app)


@pytest.fixture(name="templates")
def fixture_templates():
    return Templates()


class TemplateWrapper(typing.NamedTuple):
    name: str
    context: dict
    rendered: str
    root: str


def create_template(
    templates: Templates, tmpdir_factory, dirname: str
) -> TemplateWrapper:
    templates_dir = tmpdir_factory.mktemp(dirname)

    template = templates_dir.join("hello.html")
    template.write("<h1>Hello, {{ name }}!</h1>")

    templates.directory = str(templates_dir)

    return TemplateWrapper(
        name="hello.html",
        context={"name": "Bocadillo"},
        rendered="<h1>Hello, Bocadillo!</h1>",
        root=str(templates_dir),
    )


@pytest.fixture
def template_file(templates: Templates, tmpdir_factory) -> TemplateWrapper:
    return create_template(templates, tmpdir_factory, dirname="templates")



import pytest

from bocadillo import configure, create_client, settings


def test_basic(raw_app):
    def use_foo(app):
        @app.route("/foo")
        async def foo(req, res):
            res.text = "Foo"

    app = configure(raw_app, plugins=[use_foo])
    client = create_client(app)

    r = client.get("/foo")
    assert r.status_code == 200
    assert r.text == "Foo"


def test_use_settings(raw_app):
    def use_hello(app):
        hello_message = getattr(settings, "HELLO_MESSAGE")

        @app.route("/hello")
        async def foo(req, res):
            res.text = hello_message

    app = configure(
        raw_app, plugins=[use_hello], hello_message="Hello, plugins!"
    )
    client = create_client(app)

    r = client.get("/hello")
    assert r.status_code == 200
    assert r.text == "Hello, plugins!"


@pytest.mark.parametrize("should_use", (True, False))
def test_conditional_plugin(raw_app, should_use):
    used = False

    def use_hello(_):
        nonlocal used
        used = True

    configure(raw_app, plugins=[{use_hello: should_use}])
    assert used is should_use



import pytest

from bocadillo import configure, create_client, settings


def test_cannot_reconfigure(app):
    with pytest.raises(RuntimeError):
        configure(app)


def test_must_be_configured_to_serve(raw_app):
    @raw_app.route("/")
    async def index(req, res):
        pass

    with pytest.raises(RuntimeError) as ctx:
        with create_client(raw_app):
            pass

    assert "configure(app)" in str(ctx.value)


@pytest.mark.parametrize("positional", (True, False))
def test_settings(raw_app, positional: bool):
    class Settings:
        ONE = "one"
        _IGNORED = "ignored"
        ignored_too = "ignored too"

    args = [raw_app]
    kwargs = {"two": "two", "settings": Settings()}
    if positional:
        args.append(kwargs.pop("settings"))

    configure(*args, **kwargs)

    assert settings.ONE == "one"
    assert settings.TWO == "two"
    for name in "two", "three", "_IGNORED", "ignored_too", "IGNORED_TOO":
        assert name not in settings

    settings.ONE = 1
    assert settings.ONE == 1



from http import HTTPStatus
import typing


class HTTPError(Exception):
    """Raised when an HTTP error occurs.

    You can raise this within a view or an error handler to interrupt
    request processing.

    # Parameters
    status (int or HTTPStatus):
        the status code of the error.
    detail (any):
        extra detail information about the error. The exact rendering is
        determined by the configured error handler for `HTTPError`.

    # See Also
    - [HTTP response status codes (MDN web docs)](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status)
    """

    __slots__ = ("_status", "detail")

    def __init__(
        self, status: typing.Union[int, HTTPStatus], detail: typing.Any = ""
    ):
        if isinstance(status, int):
            status = HTTPStatus(  # pylint: disable=no-value-for-parameter
                status
            )
        else:
            assert isinstance(
                status, HTTPStatus
            ), f"Expected int or HTTPStatus, got {type(status)}"
        self._status = status
        self.detail = detail

    @property
    def status_code(self) -> int:
        """Return the HTTP error's status code, e.g. `404`."""
        return self._status.value

    @property
    def status_phrase(self) -> str:
        """Return the HTTP error's status phrase, e.g. `"Not Found"`."""
        return self._status.phrase

    @property
    def title(self) -> str:
        """Return the HTTP error's title, e.g. `"404 Not Found"`."""
        return f"{self.status_code} {self.status_phrase}"

    def as_json(self) -> dict:
        data = {"error": self.title, "status": self.status_code}
        if self.detail:
            data["detail"] = self.detail
        return data

    def __str__(self):
        return self.title



from .applications import App
from .compat import ExpectedAsync
from .config import SettingsError, configure, settings
from .errors import HTTPError
from .injection import discover_providers, provider, useprovider
from .middleware import Middleware
from .redirection import Redirect
from .request import ClientDisconnect, Request
from .response import Response
from .routing import Router
from .sse import server_event
from .staticfiles import static
from .templates import Templates
from .testing import LiveServer, create_client
from .websockets import WebSocket, WebSocketDisconnect

__version__ = "0.17.0"



import pytest

from bocadillo import App, configure, useprovider, WebSocket


@pytest.fixture
def app(raw_app: App) -> App:
    return configure(raw_app, provider_modules=["tests.data.providerconf"])


async def say_hi(req, res, hello):
    res.text = hello


class SayHi:
    async def get(self, req, res, hello):
        res.text = hello


async def say_hi_to(req, res, hello_format, who):
    res.text = hello_format.format(who=who)


class SayHiTo:
    async def get(self, req, res, hello_format, who):
        res.text = hello_format.format(who=who)


@pytest.mark.parametrize("view", (say_hi, SayHi))
def test_basic(app: App, client, view):
    app.route("/hi")(view)
    r = client.get("/hi")
    assert r.status_code == 200
    assert r.text == "Hello, providers!"


@pytest.mark.parametrize("view", (say_hi_to, SayHiTo))
def test_with_route_parameter(app: App, client, view):
    app.route("/hi/{who}")(view)
    r = client.get("/hi/peeps")
    assert r.status_code == 200
    assert r.text == "Hello, peeps!"


def test_provider_name(app: App, client):
    @app.route("/foo")
    async def say_foo(req, res, foo):
        res.text = foo

    r = client.get("/foo")
    assert r.status_code == 200
    assert r.text == "foo"


def test_websocket_clients_example(app: App, client):
    @app.websocket_route("/chat")
    async def chat(ws: WebSocket, clients: list):
        clients.add(ws)
        try:
            await ws.send_text(str(len(clients)))
            await ws.receive_text()
        finally:
            clients.remove(ws)

    @app.route("/clients")
    async def client_count(req, res, clients: list):
        res.json = {"count": len(clients)}

    with client.websocket_connect("/chat") as ws:
        assert ws.receive_text() == "1"
        r = client.get("/clients")
        assert r.status_code == 200
        assert r.json() == {"count": 1}


def test_useprovider(app: App, client):
    @app.route("/")
    @useprovider("set_called")
    async def index(req, res, spy):
        res.json = spy

    r = client.get("/")
    assert r.status_code == 200
    assert r.json()["called"] is True



import typing

from starlette.datastructures import URL
from starlette.middleware.wsgi import WSGIMiddleware
from starlette.routing import Lifespan
from starlette.websockets import WebSocketClose

from .app_types import ASGIApp, EventHandler, Receive, Scope, Send
from .config import settings
from .errors import HTTPError
from .redirection import Redirect
from .urlparse import Parser
from .views import View
from .websockets import WebSocket, WebSocketView

_V = typing.TypeVar("_V")


def _join(prefix: str, path: str) -> str:
    return prefix + path


class BaseRoute(typing.Generic[_V]):
    def matches(self, scope: dict) -> typing.Tuple[bool, dict]:
        raise NotImplementedError

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        raise NotImplementedError


class Patterned(typing.Generic[_V]):
    __slots__ = ("_pattern", "_parser", "view")

    def __init__(self, pattern: str, view: _V):
        self._parser = Parser(pattern)
        self.view = view

    @property
    def pattern(self) -> str:
        return self._parser.pattern


class HTTPRoute(BaseRoute, Patterned[View]):
    def matches(self, scope: Scope) -> typing.Tuple[bool, Scope]:
        if scope["type"] != "http":
            return False, {}
        params = self._parser.parse(scope["path"])
        if params is None:
            return False, {}
        return True, {"path_params": params}

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        req, res = scope["req"], scope["res"]
        await self.view(req, res, **scope["path_params"])


class WebSocketRoute(BaseRoute, Patterned[WebSocketView]):
    __slots__ = ("ws_kwargs",)

    def __init__(self, pattern: str, view: WebSocketView, **kwargs):
        super().__init__(pattern, view)
        self.ws_kwargs = kwargs

    def matches(self, scope: Scope) -> typing.Tuple[bool, Scope]:
        if scope["type"] != "websocket":
            return False, {}
        params = self._parser.parse(scope["path"])
        if params is None:
            return False, {}
        return True, {"path_params": params}

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        ws = WebSocket(scope, receive, send, **self.ws_kwargs)
        await self.view(ws, **scope["path_params"])


class Mount(BaseRoute):
    def __init__(self, path: str, app: ASGIApp):
        if not path.startswith("/"):
            path = "/" + path
        path = path.rstrip("/")

        self.app = app
        self.path = path
        self._parser = Parser(self.path + "/{path:path}")

    def matches(self, scope: dict) -> typing.Tuple[bool, dict]:
        path = scope["path"]

        if path == self.path + "/":
            params = {"path": ""}
        else:
            params = self._parser.parse(path)

        if params is not None:
            remaining_path = "/" + params["path"]
            matched_path = path[: -len(remaining_path)]
            child_scope = {
                "path": remaining_path,
                "root_path": scope.get("root_path", "") + matched_path,
            }
            return True, child_scope

        return False, {}

    async def __call__(self, scope, receive, send):
        try:
            await self.app(scope, receive, send)
        except TypeError:
            app = WSGIMiddleware(self.app)
            await app(scope, receive, send)


class Router:
    __slots__ = ("routes", "lifespan")

    def __init__(self):
        self.routes: typing.List[BaseRoute] = []
        self.lifespan = Lifespan()

    def add_route(self, route: BaseRoute) -> None:
        self.routes.append(route)

    def include(self, other: "Router", prefix: str = ""):
        """Include the routes of another router."""
        for route in other.routes:
            assert isinstance(route, (HTTPRoute, WebSocketRoute, Mount))
            if prefix:
                if isinstance(route, HTTPRoute):
                    route = HTTPRoute(
                        pattern=_join(prefix, route.pattern), view=route.view
                    )
                elif isinstance(route, WebSocketRoute):
                    route = WebSocketRoute(
                        pattern=_join(prefix, route.pattern),
                        view=route.view,
                        **route.ws_kwargs,
                    )
                else:
                    route = Mount(path=_join(prefix, route.path), app=route.app)
            self.add_route(route)

    def mount(self, path: str, app: ASGIApp):
        """Mount an ASGI or WSGI app at the given path."""
        return self.add_route(Mount(path, app))

    def on(self, event: str, handler: typing.Optional[EventHandler] = None):
        if handler is None:

            def register(func):
                self.lifespan.add_event_handler(event, func)
                return func

            return register

        self.lifespan.add_event_handler(event, handler)
        return handler

    def route(self, pattern: str, methods: typing.List[str] = None):
        """Register an HTTP route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.
        """

        def decorate(view: typing.Any) -> HTTPRoute:
            view = View(view, methods=methods)
            route = HTTPRoute(pattern, view)
            self.add_route(route)
            return route

        return decorate

    def websocket_route(
        self,
        pattern: str,
        *,
        auto_accept: bool = True,
        value_type: str = None,
        receive_type: str = None,
        send_type: str = None,
        caught_close_codes: typing.Tuple[int] = None,
    ):
        """Register a WebSocket route by decorating a view.

        See #::bocadillo.websockets#WebSocket for a description of keyword
        parameters.

        # Parameters
        pattern (str): an URL pattern.
        """

        def decorate(view: typing.Any) -> WebSocketRoute:
            view = WebSocketView(view)
            route = WebSocketRoute(
                pattern,
                view,
                auto_accept=auto_accept,
                value_type=value_type,
                receive_type=receive_type,
                send_type=send_type,
                caught_close_codes=caught_close_codes,
            )
            self.add_route(route)
            return route

        return decorate

    def _find_route(self, scope: dict) -> typing.Optional[BaseRoute]:
        for route in self.routes:
            matches, child_scope = route.matches(scope)
            if matches:
                scope.update(child_scope)
                return route
        return None

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        scope["send"] = send  # See: `RequestResponseMiddleware`.

        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)
            return

        route = self._find_route(scope)

        if route is not None:
            try:
                await route(scope, receive, send)
                return
            except Redirect as exc:
                scope["res"] = exc.response
                return

        try_http_redirect = (
            scope["type"] == "http"
            and not scope["path"].endswith("/")
            and redirect_trailing_slash_enabled()
        )

        if try_http_redirect:
            redirect_scope = dict(scope)
            redirect_scope["path"] += "/"
            route = self._find_route(redirect_scope)
            if route is not None:
                redirect_url = URL(scope=redirect_scope)
                scope["res"] = Redirect(str(redirect_url)).response
                return

        if scope["type"] == "websocket":
            await WebSocketClose(code=403)(receive, send)
            return

        raise HTTPError(404)


def redirect_trailing_slash_enabled() -> bool:
    return settings.get("REDIRECT_TRAILING_SLASH", True)



import typing

from .app_types import ASGIApp, ErrorHandler, Receive, Scope, Send
from .compat import check_async
from .errors import HTTPError
from .request import Request
from .response import Response

if typing.TYPE_CHECKING:  # pragma: no cover
    from .applications import App


class MiddlewareMeta(type):
    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        for callback in "before_dispatch", "after_dispatch":
            check_async(
                getattr(cls, callback),
                reason="'{name}.{callback}' must be asynchronous",
            )
        return cls


class Middleware(metaclass=MiddlewareMeta):
    """Base class for HTTP middleware classes.

    # Parameters
    inner (callable): the inner middleware that this middleware wraps.
    """

    def __init__(self, inner: ASGIApp):
        self.inner = inner

    async def before_dispatch(
        self, req: Request, res: Response
    ) -> typing.Optional[Response]:
        """Perform processing before a request is dispatched.

        If the `Response` object is returned, it will be sent and
        no further processing will be performed.
        """

    async def after_dispatch(self, req: Request, res: Response) -> None:
        """Perform processing after a request has been dispatched."""

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        assert scope["type"] == "http"

        req, res = scope["req"], scope["res"]

        before_res = await self.before_dispatch(req, res)

        if before_res is not None:
            scope["res"] = before_res
            return

        await self.inner(scope, receive, send)
        await self.after_dispatch(req, res)


class ExceptionMiddleware:
    """Handle exceptions that occur while processing requests."""

    __slots__ = ("app", "_exception_handlers")

    def __init__(
        self,
        app: ASGIApp,
        handlers: typing.Dict[typing.Type[BaseException], ErrorHandler],
    ) -> None:
        self.app = app
        self._exception_handlers = handlers

    def add_exception_handler(
        self, exception_class: typing.Type[BaseException], handler: ErrorHandler
    ) -> None:
        assert issubclass(
            exception_class, BaseException
        ), f"expected an exception class, not {type(exception_class)}"
        check_async(
            handler,
            reason=f"error handler '{handler.__name__}' must be asynchronous",
        )
        self._exception_handlers[exception_class] = handler

    def _get_exception_handler(
        self, exc: BaseException
    ) -> typing.Optional[ErrorHandler]:
        try:
            return self._exception_handlers[type(exc)]
        except KeyError:
            for cls, handler in self._exception_handlers.items():
                if isinstance(exc, cls):
                    return handler
            return None

    async def __call__(self, scope, receive, send):
        try:
            await self.app(scope, receive, send)
        except BaseException as exc:  # pylint: disable=broad-except
            if scope["type"] != "http":
                raise exc from None

            req, res = scope["req"], scope["res"]

            while exc is not None:
                handler = self._get_exception_handler(exc)
                if handler is None:
                    raise exc from None
                try:
                    await handler(req, res, exc)
                except BaseException as sub_exc:  # pylint: disable=broad-except
                    exc = sub_exc
                else:
                    exc = None


class ServerErrorMiddleware:
    """Return a 500 error response when an unhandled exception occurs."""

    __slots__ = ("app", "handler")

    def __init__(self, app: ASGIApp, handler: ErrorHandler) -> None:
        self.app = app
        self.handler = handler

    async def __call__(self, scope, receive, send):
        try:
            await self.app(scope, receive, send)
        except BaseException as exc:
            if scope["type"] != "http":
                raise exc from None
            req, res = scope["req"], scope["res"]
            await self.handler(req, res, HTTPError(500))
            raise exc from None


class RequestResponseMiddleware:
    """Make `req` and `res` available to HTTP routes and middleware.
    
    Note: this is mostly an implementation detail.
    """

    __slots__ = ("app",)

    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        response_started = False

        async def sentinel_send(message: dict):
            nonlocal response_started
            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        req = Request(scope, receive)
        scope["req"] = req
        scope["res"] = Response(req)

        try:
            await self.app(scope, receive, sentinel_send)
        finally:
            if not response_started:
                res = scope["res"]
                refreshed_send = scope.get("send", send)
                await res(scope, receive, refreshed_send)



import pytest
from starlette.responses import PlainTextResponse

from bocadillo import App, HTTPError


class ASGIMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        await self.app(scope, receive, send)


def test_asgi_middleware(app: App, client):
    params = None
    received_app = False
    called = False

    class Middleware(ASGIMiddleware):
        def __init__(self, app, **kwargs):
            super().__init__(app)
            nonlocal params, received_app
            params = kwargs
            received_app = isinstance(app, App)

        async def __call__(self, scope, receive, send):
            nonlocal called
            called = True
            await super().__call__(scope, receive, send)

    app.add_middleware(Middleware, hello="world")
    assert not received_app
    assert params == {"hello": "world"}

    @app.route("/")
    async def index(req, res):
        pass

    client.get("/")
    assert called


def test_send_response_in_middleware(app: App, client):
    class Middleware(ASGIMiddleware):
        async def __call__(self, scope, receive, send):
            response = PlainTextResponse("OK")
            await response(scope, receive, send)

    app.add_middleware(Middleware)

    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "OK"


def test_error_handling(app: App, client):
    class NotAvailableMiddleware(ASGIMiddleware):
        async def __call__(self, scope, receive, send):
            raise HTTPError(503)

    app.add_middleware(NotAvailableMiddleware)

    r = client.get("/")
    assert r.status_code == 503
    assert "Service Unavailable" in r.text


def test_middleware_called_if_routed_to_sub_app(app: App, client):
    called = False

    class Middleware:
        def __init__(self, app):
            self.app = app

        async def __call__(self, scope, receive, send):
            nonlocal called
            called = True
            await self.app(scope, receive, send)

    app.add_middleware(Middleware)

    sub = App()

    @sub.route("/")
    async def sub_index(req, res):
        pass

    app.mount("/sub", sub)

    r = client.get("/sub/")
    assert r.status_code == 200
    assert called


def test_asgi2_middleware_not_supported(app: App):
    class Middleware(ASGIMiddleware):
        def __call__(self, scope):
            pass

    with pytest.raises(ValueError) as ctx:
        app.add_middleware(Middleware)

    error = str(ctx.value).lower()
    for phrase in "asgi2", "please upgrade", "asgi3", "scope, receive, send":
        assert phrase in error



import re
import typing

PARAM_RE = re.compile(r"{}|{([a-zA-Z_:][a-zA-Z0-9_:]*)}")
WILDCARD = "{}"

CONVERTER_PATTERNS = {"path": r".*"}


def convert_part(name: str, converter: str) -> str:
    try:
        return CONVERTER_PATTERNS[converter]
    except KeyError as exc:
        raise TypeError(
            f"Unknown path converter '{converter}' "
            f"on route parameter '{name}'. "
            f"Available: {', '.join(CONVERTER_PATTERNS)}"
        ) from exc


def compile_path(pattern: str) -> typing.Tuple[typing.Pattern, str]:
    regex = "^"
    path_format = ""
    idx = 0

    for match in PARAM_RE.finditer(pattern):
        declaration, = match.groups(default="")
        name, sep, converter = declaration.partition(":")
        has_converter = sep == ":"

        regex += pattern[idx : match.start()]
        if not name:
            expr = r".*"
        else:
            part = convert_part(name, converter) if has_converter else r"[^/]+"
            expr = rf"?P<{name}>{part}"
        regex += rf"({expr})"

        path_format += pattern[idx : match.start()]
        path_format += "{%s}" % name

        idx = match.end()

    regex += pattern[idx:] + "$"
    path_format += pattern[idx:]

    return re.compile(regex), path_format


class Parser:
    def __init__(self, pattern: str):
        if pattern != WILDCARD and not pattern.startswith("/"):
            pattern = f"/{pattern}"
        self.regex, self.pattern = compile_path(pattern)

    def parse(self, value: str) -> typing.Optional[dict]:
        match = self.regex.match(value)
        if match is None:
            return None
        return match.groupdict()



import pytest

from bocadillo import App, Router, WebSocketDisconnect, configure, create_client


def test_index_returns_404_by_default(app: App, client):
    assert client.get("/").status_code == 404


def test_if_route_not_registered_then_404(app: App, client):
    assert client.get("/test").status_code == 404


def test_if_route_registered_then_not_404(app: App, client):
    @app.route("/")
    async def index(req, res):
        pass

    assert client.get("/").status_code != 404


def test_default_status_code_is_200_on_routes(app: App, client):
    @app.route("/")
    async def index(req, res):
        pass

    assert client.get("/").status_code == 200


def test_leading_slash_is_added_if_not_present(app: App, client):
    @app.route("foo")
    async def index(req, res):
        pass

    assert client.get("/foo").status_code == 200


def test_trailing_slash_not_redirected(app, client):
    @app.route("/foo")
    async def foo(req, res):
        pass

    assert client.get("/foo/").status_code == 404


@pytest.mark.parametrize("redirect", (None, True, False))
def test_redirect_trailing_slash(raw_app, redirect):
    settings = {}
    if redirect is not None:
        settings["redirect_trailing_slash"] = redirect

    app = configure(raw_app, **settings)
    client = create_client(app)

    @app.route("/foo/")
    async def foo(req, res):
        pass

    r = client.get("/foo", allow_redirects=False)

    if redirect or redirect is None:
        assert r.status_code == 302
    else:
        assert r.status_code == 404


@pytest.mark.parametrize(
    "path, status",
    [
        ("/", 404),
        ("/1", 404),
        ("/tacos", 200),
        ("/tacos/", 200),
        ("/tacos/1", 200),
        ("/tacos/gluten-free", 200),
    ],
)
def test_include_router_http(app, client, path, status):
    router = Router()

    @router.route("/")
    async def get_tacos(req, res):
        pass

    @router.route("/{pk}")
    async def get_taco(req, res, pk):
        pass

    # Starts with same prefix than router, but should not be shadowed by it.
    @app.route("/tacos/gluten-free")
    async def gluten_free_tacos(req, res):
        pass

    app.include_router(router, prefix="/tacos")

    assert client.get(path).status_code == status


def test_include_router_no_prefix(app, client):
    router = Router()

    @router.route("/foo")
    async def foo(req, res):
        pass

    app.include_router(router)

    assert client.get("/foo").status_code == 200


@pytest.mark.parametrize(
    "path, success",
    [
        ("/", False),
        ("/chat", False),
        ("/chat/", False),
        ("/tacos/chat", True),
        ("/tacos/chat/", False),
    ],
)
def test_include_router_websocket(app, client, path, success):
    router = Router()

    @router.websocket_route("/chat")
    async def chat(ws):
        await ws.send("hello")

    app.include_router(router, prefix="/tacos")

    def connect():
        with client.websocket_connect(path) as ws:
            assert ws.receive_text() == "hello"

    if success:
        connect()
    else:
        with pytest.raises(WebSocketDisconnect) as ctx:
            connect()
        assert ctx.value.code == 403


@pytest.mark.parametrize(
    "path, status", [("/other/foo", 404), ("/tacos/other/foo", 200)]
)
def test_include_router_nested_apps(app, client, path, status):
    other = App()

    @other.route("/foo")
    async def foo(req, res):
        pass

    router = Router()
    router.mount("/other", other)

    app.include_router(router, prefix="/tacos")

    assert client.get(path).status_code == status



import inspect
import typing

from . import injection
from .app_types import Handler
from .compat import check_async
from .constants import ALL_HTTP_METHODS
from .converters import ViewConverter, convert_arguments
from .errors import HTTPError

MethodsParam = typing.Union[typing.List[str], all]  # type: ignore


class HandlerDoesNotExist(Exception):
    # Raised to signal that no handler exists for a requested HTTP method.
    pass


class HTTPConverter(ViewConverter):
    def get_query_params(self, args, kwargs):
        req = args[0]
        return req.query_params


def get_handlers(obj: typing.Any) -> typing.Dict[str, typing.Callable]:
    if hasattr(obj, "handle"):
        return {method: obj.handle for method in ALL_HTTP_METHODS}
    return {
        method: getattr(obj, method)
        for method in ALL_HTTP_METHODS
        if hasattr(obj, method)
    }


class View:
    """This class defines how all HTTP views are represented internally.

    ::: warning
    Views objects should not be created directly. Instead, use
    [from_handler](#from-handler) or [from_obj](#from-obj).
    :::

    HTTP methods are mapped to **handlers**, i.e. methods of a `View` object.

    The following handlers are supported:

    - `.get(req, res, **kwargs)`
    - `.post(req, res, **kwargs)`
    - `.put(req, res, **kwargs)`
    - `.patch(req, res, **kwargs)`
    - `.delete(req, res, **kwargs)`
    - `.head(req, res, **kwargs)`
    - `.options(req, res, **kwargs)`
    - `.handle(req, res, **kwargs)`

    ::: tip
    `.handle()` is special: if defined, it overrides all others.
    :::
    """

    __slots__ = (
        "get",
        "post",
        "put",
        "patch",
        "delete",
        "head",
        "options",
        "handle",
    )

    get: Handler
    post: Handler
    put: Handler
    patch: Handler
    delete: Handler
    head: Handler
    options: Handler
    handle: Handler

    def __init__(self, obj: typing.Any, methods: typing.List[str] = None):
        if isinstance(obj, View):
            raise NotImplementedError
        if inspect.isclass(obj):
            # View-like class.
            name = obj.__name__
            class_based = True
            handlers = get_handlers(obj())
        elif callable(obj):
            # Function-based view.
            name = obj.__name__
            class_based = False
            if methods is None:
                methods = ["get"]
            if methods is all:
                methods = ["handle"]
            else:
                methods = [m.lower() for m in methods]
            handlers = {method: obj for method in methods}
        else:
            # Treat as a view-like object.
            name = obj.__class__.__name__
            class_based = True
            handlers = get_handlers(obj)

        for method, handler in handlers.items():
            full_name = f"{name}.{method.lower()}" if class_based else name
            check_async(
                handler, reason=(f"view '{full_name}()' must be asynchronous")
            )

        copy_get_to_head = "get" in handlers and "head" not in handlers
        if copy_get_to_head:
            handlers["head"] = handlers["get"]

        for method, handler in handlers.items():
            handler = convert_arguments(handler, converter_class=HTTPConverter)
            handler = injection.consumer(handler)
            setattr(self, method, handler)

    async def __call__(self, req, res, **params):
        try:
            handler = getattr(self, "handle")
        except AttributeError:
            try:
                handler = getattr(self, req.method.lower())
            except AttributeError:
                raise HTTPError(405)
        await handler(req, res, **params)
